/************************************************************************
*				HPUTILS.C				*
*	    Library of downloaded HP48G object handling utilities.	*
*                    Copyright (C) 1999 Robert Tiismus                  *
*                         Robert.Tiismus@mail.ee                        *
*                       Last update:  20. Oct. 1999                     *
*                                                                       *
*  This program is free software; you can redistribute it and/or modify *
*  it under the terms of the GNU General Public License as published by *
*  the Free Software Foundation; either version 2 of the License, or    *
*  (at your option) any later version.                                  *
*                                                                       *
*  This program is distributed in the hope that it will be useful,      *
*  but WITHOUT ANY WARRANTY; without even the implied warranty of       *
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
*  GNU General Public License for more details.                         *
*                                                                       *
*  You should have received a copy of the GNU General Public License    *
*  along with this program; if not, write to the Free Software          *
*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
************************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "hputils.h"

// const union
// { char hash[];
//   char types[];
// } type_strings;

const struct
{ long obtype;
  int ( *action )( hpFILE * );
  char type_str[ 27 ];
} prologues[] = { { DOINT,      skip_std,    "Infinite Precision Integer" },
		  { DOLNGREAL,  skip_zrel,   "Infinite Precision Real"    },
		  { DOLNGCMP,   skip_zcmp,   "Infinite Precision Complex" },
		  { DOMATRIX,   skip_comp,   "Symbolic Matrix"            },
		  { DOFLASHP,   skip_flashp, "Flash Pointer"              },
		  { DOAPLET,    skip_std,    "Aplet"                      },
		  { DOMINIFONT, skip_std,    "Minifont"                   },
		  { DOBINT,     skip_bint,   "Binary Integer"             },
		  { DOREAL,     skip_real,   "Real Number"                },
		  { DOEREL,     skip_erel,   "Long Real"                  },
		  { DOCMP,      skip_cmp,    "Complex Number"             },
		  { DOECMP,     skip_ecmp,   "Long Complex"               },
		  { DOCHAR,     hpget_b,     "Character"                  },
		  { DOARRY,     skip_std,    "Array"                      },
		  { DOLNKARRY,  skip_std,    "Linked Array"               },
		  { DOCSTR,     skip_std,    "Char String"                },
		  { DOHSTR,     skip_std,    "Hex String"                 },
		  { DOLIST,     skip_comp,   "List"                       },
		  { DORRP,      skip_rrp,    "Directory"                  },
		  { DOSYMB,     skip_comp,   "Algebraic"                  },
		  { DOEXT,      skip_comp,   "Unit"                       },
		  { DOTAG,      skip_tag,    "Tagged"                     },
		  { DOGROB,     skip_std,    "Graphic"                    },
		  { DOLIB,      skip_std,    "Library"                    },
		  { DOBAK,      skip_std,    "Backup"                     },
		  { DOEXT0,     skip_std,    "Library Data"               },
		  { DOACPTR,    skip_acptr,  "Access Pointer"             },
		  { DOEXT2,     skip_std,    "Font (HP49)"                },
		  { DOEXT3,     skip_std,    "External3"                  },
		  { DOEXT4,     skip_std,    "External4"                  },
		  { DOCOL,      skip_comp,   "Program"                    },
		  { DOCODE,     skip_std,    "Code"                       },
		  { DOIDNT,     skip_id,     "Global Name"                },
		  { DOLAM,      skip_lam,    "Local Name"                 },
		  { DOROMP,     skip_xlib,   "XLIB Name"                  } };

/** OBTYPE.STR ****************************************************************
*  Returns the string describing the object type.                             *
*  Input:  Prologue of the object                                             *
*  Output: The object type string.                                            *
******************************************************************************/
const char *obtype_str( long object )
{ int lower = 0,
      upper = sizeof( prologues ) / sizeof( prologues[ 0 ] );
  int i = ( lower + upper - 1 ) >> 1;
  long ob_i;
  static char ptr[ 10 ];
  int cont = upper;

  do
   { ob_i = prologues[ i ].obtype;
     if( object == ob_i )
       return prologues[ i ].type_str;
     else
      { if( object < ob_i )
          upper = i;
        else
          lower = i;
        i = ( lower + upper ) >> 1;
      }
   } while( cont >>= 1 );

  sprintf( ptr, "PTR %05lX", object );

  return ptr;
}

/** SKIP_OB *******************************************************************
*  Skips object in stream ( repositions the stream pointer ).                 *
*  Input:  Object prologue, input stream positioned stright after prologue    *
*          read.                                                              *
******************************************************************************/
int skip_ob( long object, hpFILE *in )
{ int lower = 0,
      upper = sizeof( prologues ) / sizeof( prologues[ 0 ] );
  int i = ( lower + upper - 1 ) >> 1;
  long ob_i;
  int cont = upper;

  do
   { ob_i = prologues[ i ].obtype;
     if( object == ob_i )
       return ( *prologues[ i ].action )( in );
     else
      { if( object < ob_i )
          upper = i;
        else
          lower = i;
        i = ( lower + upper ) >> 1;
      }
   } while( cont >>= 1 );

  return 0;
}

/******************************************************************************
*                        Object skipping routines.                            *
******************************************************************************/

int skip_zrel( hpFILE *in )
{
  skip_std( in );
  return skip_std( in );
}

int skip_zcmp( hpFILE *in )
{
  skip_zrel( in );
  return skip_zrel( in );
}

int skip_flashp( hpFILE *in )
{
  hpget_x( in );
  hpget_b( in );
  if( hpget_b( in ) == ( long )EOF )
    return EOF;
  return 0;
}

int skip_bint( hpFILE *in )
{
  if( hpget_a( in ) == ( long )EOF )
    return EOF;
  return 0;
}

int skip_real( hpFILE *in )
{
  if( hpget_w( in ) == NULL )
    return EOF;
  return 0;
}

int skip_erel( hpFILE *in )
{
  hpget_a( in );
  if( hpget_w( in ) == NULL )
    return EOF;
  return 0;
}

int skip_cmp( hpFILE *in )
{
  skip_real( in );
  return skip_real( in );
}

int skip_ecmp( hpFILE *in )
{
  skip_erel( in );
  return skip_erel( in );
}

int skip_rrp( hpFILE *in )
{ long offset;
  char *nam;

  hpget_x( in );	// Number of library, attached to directory
  if( ( offset = hpget_a( in ) ) == EOF )	// Offset of last object
    return EOF;
  if( !offset )		// Empty directory?
    return 0;
  if( hpseek( in, offset - 5L, SEEK_CUR ) )	// Seek to last object
    return EOF;
  if( ( nam = hpget_name( in ) ) == NULL )	// Skip last object name
    return EOF;
  if( strlen( nam ) )
    if( hpget_b( in ) == EOF )
      return EOF;
  return skip_ob( hpget_a( in ), in );		// Skip last object
}

int skip_tag( hpFILE *in )
{
 if( hpget_name( in ) == NULL )
   return EOF;
 return skip_ob( hpget_a( in ), in );
}

int skip_acptr( hpFILE *in )
{
  hpget_a( in );
  if( hpget_a( in ) == EOF )
    return EOF;
  return 0;
}

int skip_id( hpFILE *in )
{
  if( hpget_name( in ) == NULL )
    return EOF;
  return 0;
}

int skip_lam( hpFILE *in )
{
  if( hpget_name( in ) == NULL )
    return EOF;
  return 0;
}

int skip_xlib( hpFILE *in )
{
  hpget_x( in );
  return hpget_x( in );
}

int skip_std( hpFILE *in )
{ long offset;

  if( ( offset = hpget_a( in ) ) == EOF )
    return EOF;
  if( hpseek( in, offset - 5, SEEK_CUR ) )
    return EOF;
  return 0;
}

int skip_comp( hpFILE *in )
{ long obtype;

  while( ( obtype = hpget_a( in ) ) != SEMI )
   { if( obtype == EOF )
       return EOF;
     if( skip_ob( obtype, in ) == EOF )
       return EOF;
   }
  return 0;
}

/******************************************************************************
*               Routines to ease reading HP48G stream                         *
******************************************************************************/

hpFILE *hp_fopen( const char *name, const char *mode )
{ FILE *f;
  hpFILE *hf;

  if( ( f = fopen( name, mode ) ) == NULL )
    return NULL;

  if( ( hf = ( hpFILE * )malloc( sizeof( hpFILE ) ) ) == NULL )
   { fprintf( stderr, "ERROR: Not enough memory\n" );
     exit( 1 );
   }

  hf->stream = f;
  hf->odd = 1;
  hf->c = 0;
  hf->last_op = N_OP;
  hf->type = RF;
  
  if( strchr( mode, '+' ) != NULL )
    hf->type = RWF;
  else
    if( mode[ 0 ] == 'w' || mode[ 0 ] == 'a' )
      hf->type = WF;

  return hf;
}

int getnib( hpFILE *in )
{
  in->last_op = R_OP;
  if( in->odd = !in->odd )
   { if( in->c == EOF )
       return EOF;
     return ( in->c >> 4 ) & 0xF;
   }
  else
   { if( ( in->c = getc( in->stream ) ) != EOF )
       return in->c & 0xF;
     else
       return EOF;
   }
}

int hpget_b( hpFILE *in )
{ int c1, c2;

  c1 = getnib( in );
  if( ( c2 = getnib( in ) ) == EOF )
    return c2;

  return c1 | ( c2 << 4 );
}

int hpget_x( hpFILE *in )
{ int c1, c2;

  c1 = getnib( in ) | ( getnib( in ) << 4 );
  if( ( c2 = getnib( in ) ) == EOF )
    return EOF;

  return c1 | ( c2 << 8 );
}

long hpget_a( hpFILE *in )
{ int i, c;
  long result = 0;

  for( i = 0; i < 5; i++ )
   { c = getnib( in );
     result = result | ( ( long )c << ( i << 2 ) );
   }

  if( c == EOF )
    return ( long )c;
  return result;
}

hpword *hpget_w( hpFILE *in )
{ int i, c;
  static hpword word;

  word.hw[ 0 ] = word.hw[ 1 ] = 0L;
  for( i = 0; i < 16; i++ )
   { c = getnib( in );
     word.bytes[ i >> 1 ] |= ( unsigned char )c << ( ( i % 2 ) << 2 );
   }

  if( c == EOF )
    return NULL;
  return &word;
}

char *hpget_name( hpFILE *in )
{ static char namebuff[ 256 + 1 ];
  int namelen, i;

  if( ( namelen = hpget_b( in ) ) == EOF )
    return NULL;
  namebuff[ 0 ] = 0;
  if( namelen )
   { for( i = 0 ; i < namelen; i++ )
       if( ( namebuff[ i ] = hpget_b( in ) ) == EOF )
         return NULL;
     namebuff[ namelen ] = 0;
   }

  return namebuff;
}

unsigned do_crc( hpFILE *in, long count )
{ unsigned crc = 0;
  long i;

  for( i = 0; i < count; i++ )
    crc = ( crc >> 4 ) ^ ( ( ( crc & 0xF ) ^ getnib( in ) ) * 0x1081 );

  return crc;
}

/******************************************************************************
*               Routines to ease writing HP48G stream                         *
******************************************************************************/

void putnib( int nib, hpFILE *out )
{
  out->last_op = W_OP;
  if( out->odd = !out->odd )
   { out->c = ( out->c & 0xF ) | ( ( nib & 0xF ) << 4 );
     putc( out->c, out->stream );
   }
  else
    out->c = nib & 0xF;
}

void hpput_b( int byte, hpFILE *out )
{
  putnib( byte, out );
  putnib( byte >> 4, out );
}

void hpput_x( int ext, hpFILE *out )
{ int i;

  for( i = 0; i < 3; i++ )
    putnib( ext >> ( i << 2 ), out );
}

void hpput_a( long addr, hpFILE *out )
{ int i;

  for( i = 0; i < 5; i++ )
    putnib( ( int )( addr >> ( i << 2 ) ), out );
}

void hpput_w( hpword word, hpFILE *out )
{ int i;

  for( i = 0; i < 16; i++ )
    putnib( ( int )( word.bytes[ i >> 2 ]  >> ( ( i % 2 ) << 2 ) ), out );
}

int hp_fclose( hpFILE *out )
{ int result;

  if( out->last_op == W_OP )
    switch( out->type )
     { case WF:
         if( !out->odd )
           putnib( 0, out );
         break;
       case RWF:
         if( !out->odd )
           hpseek( out, 0, SEEK_CUR );
     }

  result = fclose( out->stream );
  free( out );
  
  return result;
}

/******************************************************************************
*               Routines to ease handling HP48G stream                        *
******************************************************************************/

int hpseek( hpFILE *in, long offset, int whence )
{ long local_offs;
  int result, c;

  if( in->last_op == W_OP )
   { switch( in->type )
       { case WF:
           if( !in->odd )
            { putnib( 0, in );
	      fseek( in->stream, -1, SEEK_CUR );
	      in->odd = !in->odd;
	    }
           break;
         case RWF:
           if( !in->odd )
            { fseek( in->stream, 0, SEEK_CUR );
              c = ( getc( in->stream ) >> 4 ) & 0xF;
              fseek( in->stream, -1, SEEK_CUR );
              putnib( c, in );
              fseek( in->stream, -1, SEEK_CUR );
	      in->odd = !in->odd;
            }
       }
   }

  in->last_op = N_OP;

  switch( whence )
   { case SEEK_CUR:
       local_offs = offset / 2;
       if( !in->odd && !( offset % 2 ) )
         --local_offs;
       result = fseek( in->stream, local_offs, SEEK_CUR );
       in->odd = in->odd ^ ( offset % 2 );
       break;
     case SEEK_SET:
       local_offs = offset / 2;
       result = fseek( in->stream, local_offs, SEEK_SET );
       in->odd = !( offset % 2 );
   }

  if( !( in->odd || result ) )
   { in->odd = !in->odd;
     if( getnib( in ) == EOF )
       result = -1;
   }

  return result;
}

void hpsync_w( hpFILE *hf )
{
  if( hf->last_op == R_OP )
    fseek( hf->stream, -1, SEEK_CUR );
}

long hptell( hpFILE *in )
{ long result;

  if( ( result = ftell( in->stream ) ) != -1L )
   { result *= 2;
     if( !in->odd )
       --result;
   }

  return result;
}

